// To do: next major: remove this file?
export type {
  // Used in `unist-util-visit`:
  VisitorResult,

  // Documented:
  Visitor,
  BuildVisitor
} from './index.js'
export type {
  // Used in `unist-util-visit`:
  Matches,
  InclusiveDescendant
} from './lib/complex-types.js'
